﻿var fs = require('fs');
var Promise = require('promise');
var jsMapper = require("./js_diff/index.js").javascriptHandler;
var jsMapperV2 = require("./js_diff/index.js").javascriptHandlerV2;
var mapSingleFileInMemory = require("./js_diff/index.js").mapSingleFileInMemory;
var filesMapping = require("./js_diff/files-mapping");
var consoleFileLogger = require("./sl-logger.js").consoleFileLogger;
var version = require("../package.json").version;
var slNodeCover = require('sl-node-cover');
var SlNeedle = require('./sl-needle.js').SlNeedle;
var NewDiffService = require("./new-diff-service.js").NewDiffService;
var md5 = require('md5');
var utils = require('./utils.js');
var os = require("os");

function BuildDiffProcess(cfg, diffService, sourceConrolProviders, parsedDependencies, logger) {

    var log = logger ? logger.child({
        className: "BuildDiffProcess"
    }) : consoleFileLogger("BuildDiffProcess");

    function createBuildDiffPromiseChain(buildArguments) {
        var handlerParams = {
            buildArguments: buildArguments,
            cfg: cfg,
            finalOutput: {
                "meta": {
                    "generated": new Date().valueOf(),
                    "build": buildArguments.build,
                    "commit": buildArguments.commit,
                    "customerId": cfg.customerId,
                    "appName": buildArguments.appname,
                    //"technologies": "nodejs",
                    "technology": "nodejs",
                    "branch": buildArguments.branch,
                    "environment": {
                        agentVersion: version,
                        agentType: "nodeJsBuildScanner",
                        processArgv:process.argv,
                        machineName: os.hostname(),
                        platform: os.platform(),
                        os: os.type(),
                        osVersion: os.release(),
                        arch: os.arch(),
                        processId: process.pid,
                        ipAddress: clone(os.networkInterfaces()),
                        dependencies: clone(process.versions),
                        runtime: process.version
                    },
                    "authors": buildArguments.author,
                    "logsUrl": buildArguments.logsUrl,
                    "jobName": buildArguments.jobName,
                    "moduleName":  buildArguments.uniqueModuleId, //Added for backward compatibility incase the server needs it.
                    "uniqueModuleId": buildArguments.uniqueModuleId,
                    "configurationData": buildArguments,
                    "scmProvider": buildArguments.scmProvider,
                    "scmVersion": buildArguments.scmVersion,
                    "scmBaseUrl": buildArguments.scmBaseUrl
                },
                version: buildArguments.useNewBuildScannerProtocol ? "2.0" : "1.0",
                files: [],
                dependencies: parsedDependencies || []
            }
        };

        var scmInstance;

        if (buildArguments.workspacepath) {
            var scmType = sourceConrolProviders[buildArguments.scm];
            if (scmType) {
                scmInstance = new scmType(cfg, buildArguments.workspacePath);
                handlerParams.finalOutput.meta.scm = buildArguments.scm;
            }
        }

        if (!buildArguments.commit && scmInstance && !scmInstance.detectCommitVersion) {
            return Promise.reject("'commit' command-line argument is required, since automatic detection by the " + buildArguments.scm + " SCM is not supported");
        }

        var promiseChain = Promise.resolve();

        if (buildArguments.workspacepath) {

            //2. If there is a mapping, get the diffs
            if (scmInstance) {
                //2.1 Get repository url from the SCM
                promiseChain = promiseChain.then(function (mapping) {
                    return getRepositoryUrl(handlerParams, scmInstance);
                });

                //2.2 Get all contributors from the scm
                if(buildArguments.sendContributors){
                    promiseChain = promiseChain.then(function(mapping){
                        return getContributors(handlerParams, scmInstance);
                    });
                }

                //2.3 Get branch history from the SCM
                promiseChain = promiseChain.then(function (mapping) {
                    return loadBranchHistory(handlerParams, scmInstance);
                });

                //2.4 Get all commits effected each file
                if(buildArguments.sendContributors){
                    promiseChain= promiseChain.then(function(mapping){
                        return getCommitsPerFile(handlerParams, scmInstance);
                    });
                }
                //2.5 if commit was not specified, automatically detect it
                if (!buildArguments.commit) {
                    promiseChain = promiseChain.then(function () {
                        return scmInstance.detectCommitVersion(handlerParams, log).then(function (commit) {
                            if (commit) {
                                handlerParams.finalOutput.meta.commit = commit;
                            }
                        });
                    });
                };
            }

            //3. Call JS mapping handler
            promiseChain = promiseChain.then(function () {
                 if(buildArguments.useNewBuildScannerProtocol){
                     var filesMapper = new filesMapping(handlerParams, log);
                     return filesMapper.process();
                 }
                return jsMapper(handlerParams, log);
            });
        };

        //4. Submit the final output to the server
        promiseChain = promiseChain.then(function () {
            if (buildArguments.instrumentationOnly) {
                return Promise.resolve();
            } else {                
                //fs.writeFileSync("buildMapping.json", JSON.stringify(handlerParams.finalOutput,null,'\t'));
                //return Promise.resolve();
                return diffService.submitMapping(handlerParams.finalOutput);
            }
        });

        //5. Create instrumented browser files.
        if (buildArguments.instrumentForBrowsers) {
            promiseChain = promiseChain.then(function () {
                if (handlerParams.finalOutput && handlerParams.finalOutput.files) {
                    delete handlerParams.finalOutput.files;
                }
                var promise = new Promise(function (resolve, reject) {
                    var resolveWithoutHash = buildArguments.useNewBuildScannerProtocol || buildArguments.isUseBuildMappingV3 || false;
                    var delayShutdownInSeconds =buildArguments.delayShutdownInSeconds || 30;
                    var downloadAgent = buildArguments.downloadAgent != null ? buildArguments.downloadAgent : true;
                    slNodeCover.utils.runInstrumentation({
                        appName: buildArguments.appname,
                        build: buildArguments.build,
                        customerId: cfg.customerId,
                        branch: buildArguments.branch,
                        sourceRoot: buildArguments.workspacepath,
                        files: handlerParams.processedFiles, //Will be filled by jsMapper in step 3
                        outputPath: buildArguments.outputpath,
                        copyAllFilesToOutput: buildArguments.copyAllFilesToOutput !== false,
                        buildsessionid: buildArguments.buildsessionid,
                        esModules: buildArguments["es-modules"],
                        server: handlerParams.cfg.server,
                        token: cfg.token,
                        instrumentationType: "browser",
                        resolveWithoutHash: resolveWithoutHash,
                        delayShutdownInSeconds: delayShutdownInSeconds,
                        downloadAgent: downloadAgent,
                        workspacepath: buildArguments.workspacepath
                    }, function (err) {
                        if (err)
                            reject(err);
                        else
                            return resolve();
                    });
                });
                return promise;
            });
        }

        promiseChain = promiseChain.then(function () {
            return Promise.resolve();
        }).catch(function (err) {            
            return Promise.reject(err);
        });

        return promiseChain;
    }

    this.run = function run(buildArguments) {
        return createBuildDiffPromiseChain(buildArguments);
    }

    function loadBranchHistory(handlerParams, scmInstance) {
        return scmInstance.getBranchHistory(handlerParams, log).then(function (gitLog) {
            if (gitLog && gitLog.commitHistory && gitLog.commitHistory.length && gitLog.commitLog && gitLog.commitLog.length) {
                log.info('Branch history loaded (' + gitLog.length + ' commits)');
                handlerParams.finalOutput.meta.history = gitLog.commitHistory;
                handlerParams.finalOutput.meta.commitLog = gitLog.commitLog;
            } else {
                logBranchHistoryNotFound(gitLog, log);
            }
            if (gitLog && gitLog.commitToIndexMap) {
                handlerParams.commitToIndexMap = gitLog.commitToIndexMap;
            } else {
                log.error("Commit log index map not laded")
            }
            return gitLog;
        });
    }

    function getRepositoryUrl(handlerParams, scmInstance) {
        return scmInstance.getRepositoryUrl(handlerParams, log).then(function (repositoryUrl) {
            if (repositoryUrl) {
                log.info('Repository Url loaded (' + repositoryUrl + ')');
                handlerParams.finalOutput.meta.repositoryUrl = repositoryUrl;
            }
            return repositoryUrl;
        });
    }

    function getContributors(handlerParams, scmInstance) {
        return scmInstance.getContributors(handlerParams, log).then(function (contributorsLog) {
            if (contributorsLog && contributorsLog.contributors && contributorsLog.emailToIndexMap) {
                log.info('Contributors list loaded (' + contributorsLog.contributors.length + ' contributors)');
                handlerParams.finalOutput.meta.contributors = contributorsLog.contributors;
                handlerParams.emailToIndexMap = contributorsLog.emailToIndexMap;
            }else{
                logContributorsNotFound(contributorsLog, log);
            }
            return contributorsLog;
        })
    }

    function getCommitsPerFile(handlerParams, scmInstance){
        return scmInstance.getCommitsPerFile(handlerParams, log).then(function(fileToCommitsMap){
            if(fileToCommitsMap != null){
                log.info('File commit map loaded (' + fileToCommitsMap.length + ' records)');
                handlerParams.fileToCommitsMap = fileToCommitsMap;
            }else{
                log.error("Commits map is null or undefined.")
            }
            return fileToCommitsMap;
        })
    }
}

function InstrumentingProxyTransformingProcess(diffService) {
    this.run = function run(options, logger) {
        var logger = logger || new consoleFileLogger("build-scanner-proxy");
        var needle = new SlNeedle(options.token);
        var diffService = new NewDiffService({
            proxy: options.proxy,
            isValidToken: options.isValidToken,
            server: options.server
        }, needle, logger);

        var mapFileOpts = {
            es6Modules: options.es6Modules,
            prefixesOfExcludedFiles: options.prefixesOfExcludedFiles,
            useBabylon: options.useBabylon
            };

        var fileSig = mapSingleFileInMemory({
            filename: options.filename,
            fullFilename: options.originalUrl,
            buffer: options.originalFileContent,
            usebranchcoverage: true
        }, mapFileOpts, logger);

        var moduleName = md5(options.originalUrl); //Uncomment this to include hash in module name, but this ruin trends since modules will never be the same: /* + "#" + fileSig.hash */

        var finalOutput = {
            "meta": {
                "generated": new Date().valueOf(),
                "build": options.buildName,
                "commit": "",
                "customerId": options.customerId,
                "appName": options.appName,
                "moduleName": moduleName,
                "technology": "nodejs",
                "branch": options.branchName,
                "environment": {
                    agentVersion: version,
                    agentType: "nodeJsBuildScanner",
                    processArgv:process.argv,
                    machineName: os.hostname(),
                    platform: os.platform(),
                    os: os.type(),
                    osVersion: os.release(),
                    arch: os.arch(),
                    processId: process.pid,
                    ipAddress: clone(os.networkInterfaces()),
                    dependencies: clone(process.versions),
                    runtime: process.version
                },
                "configurationData": options
            },
            files: [fileSig],
            dependencies: []
        };

        var promiseChain = Promise.resolve();
        promiseChain = promiseChain.then(function () {
            logger.info("Submitting mapping...");
            return diffService.submitMapping(finalOutput);
        }).then(function () {
            var promise = new Promise(function (resolve, reject) {
                var inMemoryFile = {
                    filename: options.filename,
                    content: options.originalFileContent,
                    instrumented: null
                };

                var resolveWithoutHash = options.useNewBuildScannerProtocol || options.isUseBuildMappingV3 || false;
                var delayShutdownInSeconds =options.delayShutdownInSeconds || 30;
                var downloadAgent = options.downloadAgent != null ? options.downloadAgent : true;
                slNodeCover.utils.runInstrumentation({
                    appName: options.appName,
                    build: options.buildName,
                    customerId: options.customerId,
                    branch: options.branchName,
                    files: [],
                    inMemoryFile: inMemoryFile,
                    additionalFilesToCopy: [],
                    outputPath: null,
                    buildsessionid: options.buildsessionid,
                    server: options.server,
                    token: options.token,
                    instrumentationType: "browser",
                    resolveWithoutHash: resolveWithoutHash,
                    delayShutdownInSeconds: delayShutdownInSeconds,
                    downloadAgent: downloadAgent,
                    workspacepath: options.workspacepath

                }, function (err) {
                    if (err) {
                        logger.error("Error: ", err);
                        reject(err);
                    } else {
                        logger.info("Done (success)");
                        return resolve(inMemoryFile.instrumented);
                    }
                });
            });
            return promise;
        });

        return promiseChain;
    }
}

function logContributorsNotFound(contributorsLog, log) {
    var fieldsMap = {
        contributorsLog: !contributorsLog,
        contributors: contributorsLog && !contributorsLog.contributors,
        emailToIndexMap: contributorsLog && !contributorsLog.emailToIndexMap
    };
    var missing = utils.getMissingProperties(fieldsMap)
    log.error("Contributors list not loaded, missing fields: " + missing);
}

function logBranchHistoryNotFound(gitLog, log) {
    var fieldsMap = {
        gitLog: !gitLog,
        commitHistory: gitLog && !gitLog.commitHistory,
        commitHistoryLength: gitLog && gitLog.commitHistory && !gitLog.commitHistory.length,
        commitLog: gitLog && !gitLog.commitLog,
        commitLogLength: gitLog && gitLog.commitLog && !gitLog.commitLog.length
    };
    var missing = utils.getMissingProperties(fieldsMap)
    log.error("Branch history not loaded, missing fields: " + missing);
}

function clone(o) {
    //if (!o) return o; //Unreachable
    return JSON.parse(JSON.stringify(o));
}

module.exports.InstrumentingProxyTransformingProcess = InstrumentingProxyTransformingProcess;
module.exports.BuildDiffProcess = BuildDiffProcess;
module.exports.DiffService = require("./new-diff-service.js").NewDiffService;
module.exports.sourceControlProviders = require("./source-control-providers.js");