var fileAndFolderUtils = require('../file-and-folder-utils');
var slIgnoreWrapper = require('../sl-ignore-wrapper');
var glob = require("glob");
var path = require('path');
var fs = require('fs');
var ActualFile = require('../file-system/actual-file');
var utils = require('../utils');
var consoleFileLogger = require("../sl-logger.js").consoleFileLogger;
var fileSignature = require('./file-signature.js');
var methodKeyCreator = require('./method-key.js');
var branchKeyCreator = require('./branch-key.js');
var uuid = require('uuid');
var InMemoryFile = require('../file-system/in-memory-file');

var SCAN_IGNORE_FILE_NAME = ".slignore";
var GENERATED_IGNORE_FILE_NAME = ".slignore.generated";
var BRANCHES_KEY = "branches";
var BRANCH_ELEMENT_TYPE = "branch";
var METHODS_KEY = "methods";
var METHOD_ELEMENT_TYPE = "method";
function FilesMapping(handlerParams, logger){
    this.handlerParams = handlerParams || {};
    this.options = this.handlerParams.buildArguments;
    this.finalOutput = this.handlerParams.finalOutput;
    this.fileNames = [];
    this.filesLookup = {};
    this.fileSignatures = [];
    this.handlerParams.processedFiles = [];
    this.erroredFiles = [];
    this.finalFilesData =[];
    this.logger = logger;
    this.log = this._createLogger(logger, "FileData");
    this.slIgnoreScan = this._createSlIgnore(SCAN_IGNORE_FILE_NAME);
    this.slIgnoreGenerated = this._createSlIgnore(GENERATED_IGNORE_FILE_NAME);
}

FilesMapping.prototype.process = function () {
    var that = this
    var promise = new Promise(function (resolve) {
        that._populateFilesArray();
        that.fileNames.forEach(function (filename, fileIdx) {
            that._fillFilesData(filename,  fileIdx);
        });
        that.filesLookup ={};
        that.fileSignatures.forEach(function (file) {
            that._fillMethodsData( file);
            that._fillBranchesData(file);
        });
        if(that.erroredFiles.length){
            var message = that._getErroredFilesMessage();
            that.log.warn(message);
            return resolve(false)
        }
        that.finalOutput.files = that.finalFilesData;
        return resolve(true);
    });
    return promise;
}

FilesMapping.prototype._populateFilesArray = function () {
    var that = this
    that.options.includedFiles.forEach(function (inclusionPattern) {
        var matchedFiles = glob.sync(inclusionPattern, {
            cwd: that.options.workspacepath,
            ignore: that.options.excludedFiles
        });
        matchedFiles.forEach(function (matchedFile) {
            if (!that.filesLookup[matchedFile]) {
                that.filesLookup[matchedFile] = true;
                that.fileNames.push(matchedFile);
            }
        });
    });
}

FilesMapping.prototype._fillFilesData = function (filename, fileIdx) {
    var fullFilename = path.resolve(this.options.workspacepath, filename);
    var shouldSkipCurrFile = this._shouldSkipFile(fullFilename, filename, fileIdx);
    if (shouldSkipCurrFile)
        return;
    this.handlerParams.processedFiles.push(filename);
    this.log.info('Processing [' + (fileIdx + 1) + '/' + this.fileNames.length + '] ' + fullFilename);
    var fileSig = this._createFileSignature(filename, fullFilename);
    if (fileSig) {
        fileSig.methods = this._getIncludedMethods(fileSig);
        fileSig.branches = this._getIncludedBranches(fileSig);
        this.fileSignatures.push(fileSig);
    }
    else {
        this.erroredFiles.push(filename);
    }
}

FilesMapping.prototype._shouldSkipFile = function (fullFilename, filename, fileIdx) {
    var that = this;
    var filePath = fs.statSync(fullFilename);
    if (!filePath.isFile()) {
        that.log.info(filename + " is not a file.")
        return true;
    }
    // If the file is filtered by the .slignore file:
    if (that.slIgnoreScan.denies(filename)) {
        that.log.info('Excluded [' + (fileIdx + 1) + '/' + that.fileNames.length + '] ' + fullFilename);
        // Ignored by .slignore, skip it.
        return true;
    }
    return false;
}

FilesMapping.prototype._createFileSignature = function (filename, fullFilename) {
    var useBranchCoverage = this.handlerParams.cfg.useBranchCoverage || this.options.usebranchcoverage;
    var fileObject = new ActualFile(filename, fullFilename);
    var fileSig = this._generateNewMappingForFile(fileObject, useBranchCoverage);
    return fileSig;
}

FilesMapping.prototype._generateNewMappingForFile = function (fileObject, isBranchCoverage) {
    var that = this;
    var filename = fileObject.getFilename();
    var fullname = fileObject.getFullFilename();
    filename = utils.adjustPathSlashes(filename);
    var fileSig = {
        filename: filename,
        methods: []
    };
    var oldContent = fileObject.getContent();
    var sourceMaps = fileObject.readSourceMaps(this.logger);
    var sig = fileSignature.generateFileSignature(fullname, filename, oldContent, sourceMaps, isBranchCoverage, that.options, that.logger);
    if (!sig || sig.error) {
        return null;
    }
    fileSig.hash = sig.module.hash;
    fileSig.hasSourceMaps = !!sourceMaps;

    sig.methods.forEach(function (method) {
        var wholeKey = [methodKeyCreator(method.start, method.name, method.isAnonymous), method.hash, new Date().valueOf()].join('|');
        var newId = uuid(wholeKey);
        var methodObject = that._createMethodObject(method, newId);
        fileSig.methods.push(methodObject)
    });
    return fileSig;
}

FilesMapping.prototype._createLogger = function (logger, methodName) {
    var log = logger ? logger.child({
        methodName: methodName
    }) : consoleFileLogger(methodName);
    return log;
}

FilesMapping.prototype._createMethodObject = function (method, newId) {
    return {
        id: newId,
        isAnonymous: method.isAnonymous,
        hash: method.hash,
        sigHash: method.sigHash,
        type: method.type,
        position: method.start,
        endPosition: method.end,
        name: method.name,
        uniqueName: method.uniqueName,
        filename: method.originalFilename,
        originalFilename: method.filename,
        originalStart: method.originalStart,
        originalEnd: method.originalEnd,
        originalName: method.originalName,
        branches: method.branches
    }
}

FilesMapping.prototype._createBranchObject = function (uniqueId, branch) {
    var branchData = {
        uniqueId: uniqueId,
        filename: branch.filename,
        type: branch.type,
        position: branch.position,
        endPosition: branch.endPosition,
        hash: branch.hash,
        index: branch.index
    }
    return branchData;
}

FilesMapping.prototype._getIncludedMethods = function (fileSig) {
    if (!fileSig.methods)
        return [];
    return this._getIncludeElements(fileSig, METHODS_KEY, METHOD_ELEMENT_TYPE);;
}



FilesMapping.prototype._getIncludedBranches =  function (fileSig) {
    if (!fileSig.branches)
        return []
    return this._getIncludeElements(fileSig, BRANCHES_KEY, BRANCH_ELEMENT_TYPE);

}

FilesMapping.prototype._getIncludeElements = function (fileSig, key, elementType) {
    var that = this;
    var includedElements = [];
    fileSig[key].forEach(function (element) {
        if (element.originalFilename) {
            that._handleGeneratedElement(element, includedElements, elementType);
        } else {
            includedElements.push(element);
        }
    });
    return includedElements;
}

FilesMapping.prototype._handleGeneratedElement = function (element, elementsArr, elementType) {
    if (this.slIgnoreGenerated.accepts(element.originalFilename)) {
        elementsArr.push(element);
    } else {
        this.log.info('Rejecting ' + elementType + ' from file: ' + element.originalFilename);
    }
}

FilesMapping.prototype._getErroredFilesMessage = function () {
    return "The following files failed to be parsed and will not be sent to the server:\n" +
        this.erroredFiles.join('\n') +
        "\n\t(Hint: if the file use the es6 module syntax, pass the \"--es6Modules\" switch)";
}

FilesMapping.prototype._handleMethodInFile = function (method, file) {
    var trueSrcFilename = method.originalFilename || file.filename;
    // Skip method that is not mapped to any "true" source file. It's compiler generated
    if (file.hasSourceMaps && trueSrcFilename == file.filename) {
        this.log.info("method: " + method.uniqueName + ". in file: " + file.filename + ". it's a generated method.")
        return;
    }
    var methodFile = this._getCreateFile(trueSrcFilename);
    var convertedMethod = this._convertMethod(trueSrcFilename, method);
    methodFile.methods.push(convertedMethod);
    file.branches = file.branches || [];
    if (method.branches) {
        var that = this;
        method.branches.forEach(function (branch) {
            return that._extractBranches(branch, method, file);
        });
    }
}

FilesMapping.prototype._handleBranchInFile = function (branch, file) {
    var trueSrcFilename = branch.originalFilename || file.filename;
    if (file.hasSourceMaps && trueSrcFilename == file.filename) { // Skip method that is not mapped to any "true" source file. It's compiler generated
        return;
    }
    var branchFile = this._getCreateFile(trueSrcFilename);
    var uniqueId = branchKeyCreator(branch.originalFilename, branch.parentPosition, branch.index);
    var branchData = this._createBranchObject(uniqueId, branch);
    branchFile.branches.push(branchData);
}

FilesMapping.prototype._getCreateFile = function (filename) {
    if (!this.filesLookup[filename]) {
        this._handleNewFile(filename);
    }
    return this.filesLookup[filename];
}

FilesMapping.prototype._handleNewFile = function (filename) {
    var newFile = this._createNewFile(filename);
    var addCommitsArray = this._shouldAddCommitsArray()
    if (addCommitsArray) {
        newFile.commitIndexes = this.handlerParams.fileToCommitsMap[filename] || [];
    } else {
        this._logCommitsArrayNotFound();
    }
    this.filesLookup[filename] = newFile;
    this.finalFilesData.push(newFile);
}

FilesMapping.prototype._createNewFile = function (filename) {
    var newFile = {
        logicalPath: filename,
        physicalPath: filename,
        methods: [],
        lines: [],
        branches: []
    };
    return newFile;
}

FilesMapping.prototype._shouldAddCommitsArray = function() {
    return this.options && this.options.sendContributors && this.handlerParams.fileToCommitsMap;
}

FilesMapping.prototype._logCommitsArrayNotFound = function () {
    var fieldsMap = {
        opts: !this.options,
        sendContributors: this.options && !this.options.sendContributors,
        fileToCommitsMap: !this.handlerParams.fileToCommitsMap
    }
    var missing = utils.getMissingProperties(fieldsMap);
    this.log.error("Cannot add commits indexes, missing fields: " + missing);
}

FilesMapping.prototype._convertMethod = function (trueSrcFilename, method) {
    var convertedMethod = {
        uniqueId: [trueSrcFilename, method.position.join(',')].join('@'),
        displayName: method.name,
        position: method.position,
        endPosition: method.endPosition,
        hash: method.hash,
        sigHash: method.sigHash,
        meta: {} //Filled-in below
    };
    if (method.isAnonymous) {
        convertedMethod.meta.anonymous = true;
    }
    if (method.type) {
        convertedMethod.meta.type = method.type;
    }
    return convertedMethod;
}

FilesMapping.prototype._extractBranches = function (branch, method, file) {
    if (!branch){
        return;
    }
    //TODO: Hack! Currently the branches which we collect are inside method.
    //When we'll support global branches, we should get the "original file name"
    //in the same time we create the branch data object.
    branch.originalFilename = method.originalFilename;
    file.branches.push(branch);
}

FilesMapping.prototype._createSlIgnore = function (filename) {
    var workspacePath = this.options.workspacepath;
    var lookInParentFolders = !!this.options.instrumentForBrowsers;
    var slIgnoreScanFilePath = fileAndFolderUtils.findFileRecursively(filename, workspacePath, lookInParentFolders, this.logger);
    var slIgnoreScan = new slIgnoreWrapper(slIgnoreScanFilePath, this.options.outputpath, this.log);
    return slIgnoreScan;
}

FilesMapping.prototype._fillMethodsData = function (fileSig) {
    var that = this;
    if (!fileSig.methods) {
        that.log.info("file: " + fileSig.filename + ". does not containing methods")
    } else {
        fileSig.methods.forEach(function (method) {
            that._handleMethodInFile(method, fileSig);
        })
    }
}

FilesMapping.prototype._fillBranchesData = function (file) {
    var that = this;
    if (!file.branches) {
        this.log.info("file: " + file.filename + ". does not containing branches")
    } else {
        file.branches.forEach(function (branch) {
            that._handleBranchInFile(branch, file);
        });
    }
}

FilesMapping.prototype.mapSingleFileInMemory = function (singleFileParams) {
    var useBranchCoverage = singleFileParams.usebranchcoverage;
    var fileObject = new InMemoryFile(singleFileParams.filename, singleFileParams.originalFilename, singleFileParams.buffer);
    var fileSig = this._generateNewMappingForFile(fileObject, useBranchCoverage);
    return fileSig;
}

module.exports = FilesMapping;